unit Delaunay;

interface

uses Dialogs, Graphics, Forms,Classes,Windows;
//UnitCubicSpline;

//Set these as applicable
Const MaxVertices = 500000;
Const MaxTriangles = 1000000;
Const ExPtTolerance = 0.01;

//Points (Vertices)
Type dVertex = record
    x: Double;
    y: Double;
    z: double;
    ID:Integer;
    ln:Shortint; // Vorkommen in Linien
end;

//Created Triangles, vv# are the vertex pointers
Type dTriangle = record
    vv0: LongInt;
    vv1: LongInt;
    vv2: LongInt;
    PreCalc: Integer;
    xc,yc,r: Double;
end;

type TDVertex = array[0..MaxVertices] of dVertex;
type PVertex = ^TDVertex;

type TDTriangle = array[0..MaxTriangles] of dTriangle;
type PTriangle = ^TDTriangle;

type TDComplete = array [0..MaxTriangles] of Boolean;
type PComplete = ^TDComplete;

type TDEdges = array[0..2,0..MaxTriangles * 3] of LongInt;
type PEdges = ^TDEdges;

Type TPunkt=Record
     X,Y,Z:Double;
     ln:Shortint; // Vorkommen in Linien
     end;
type TLine = record
     sp: Integer;
     ep: Integer;
     AE: Char;
     end;


  type TContourSetup = record
   MinValue: Double;
   MaxValue: Double;
   Range: Double;
   Step: Double;
   NoLevels: Integer;
   Levels: array of Double;
   Points:array of TPunkt;
   Contours: array of array of TLine;
  end;

  type TLineCut = record
    ContourPt: Tpunkt;
    Found: Boolean;
    end;


type
  TDelaunay = class
  private
    { Private declarations }
    function InCircle(xp, yp, x1, y1, x2, y2, x3, y3: Double;
             var xc: Double; var yc: Double; var r: Double; j: Integer): Boolean;
    Function WhichSide(xp, yp, x1, y1, x2, y2: Double): Integer;
    Function Triangulate(nvert: Integer): Integer;
  public
    { Public declarations }
    Vertex: PVertex;
    Triangle: PTriangle;
    TempBuffer: TBitmap;
    HowMany: Integer;
    zMin,zMax:double;
    tPoints: Integer; //Variable for total number of points (vertices)
    Contour: TContourSetup;
//    TargetForm: TForm;
    constructor Create;
    destructor Destroy; override;
    procedure Mesh;
//    procedure Draw;
    procedure AddPoint(Number:Integer;x,y,z: Double);
//    procedure ClearBackPage;
//    procedure FlipBackPage;
    procedure QuickSort(var A: PVertex; Low,High: Integer);
    function LineCut(x1,y1,x2,y2,z1,z2,Level: Double): TLineCut;
    procedure InitialiseContourLevels(MinZ,MaxZ,Step:Double);
    procedure CalculateContours;
    function AddContourPoint(P:Tpunkt):Integer;
    procedure AddContourLine(Level:Integer;L1,L2:TLineCut);
    procedure SortContourLines;
 //   procedure DrawContours(Canvas:Tcanvas);

  end;

implementation

constructor TDelaunay.Create;
begin
  //Initiate total points to 1, using base 0 causes problems in the functions
  tPoints := 1;
  HowMany:=0;
//  TempBuffer:=TBitmap.Create;

  //Allocate memory for arrays
  GetMem(Vertex, sizeof(Vertex^));
  GetMem(Triangle, sizeof(Triangle^));
end;

destructor TDelaunay.Destroy;
begin
  //Free memory for arrays
  FreeMem(Vertex, sizeof(Vertex^));
  FreeMem(Triangle, sizeof(Triangle^));
end;

function TDelaunay.InCircle(xp, yp, x1, y1, x2, y2, x3, y3: Double;
                            var xc: Double; var yc: Double;
                            var r: Double; j: Integer): Boolean;
//Return TRUE if the point (xp,yp) lies inside the circumcircle
//made up by points (x1,y1) (x2,y2) (x3,y3)
//The circumcircle centre is returned in (xc,yc) and the radius r
//NOTE: A point on the edge is inside the circumcircle
var
  eps: Double;
  m1: Double;
  m2: Double;
  mx1: Double;
  mx2: Double;
  my1: Double;
  my2: Double;
  dx: Double;
  dy: Double;
  rsqr: Double;
  drsqr: Double;
begin
  eps:= 0.000001;
  InCircle := False;

  //Check if xc,yc and r have already been calculated
  if  Triangle^[j].PreCalc=1 then
  begin
    xc := Triangle^[j].xc;
    yc := Triangle^[j].yc;
    r  := Triangle^[j].r;
    rsqr := r*r;
    dx := xp - xc;
    dy := yp - yc;
    drsqr := dx * dx + dy * dy;
  end
  else
  begin
    If (Abs(y1 - y2) < eps) And (Abs(y2 - y3) < eps) Then
    begin
      ShowMessage('INCIRCUM - F - Points are coincident !!');
      Exit;
    end;

    If Abs(y2 - y1) < eps Then
    begin
      m2 := -(x3 - x2) / (y3 - y2);
      mx2 := (x2 + x3) / 2;
      my2 := (y2 + y3) / 2;
      xc := (x2 + x1) / 2;
      yc := m2 * (xc - mx2) + my2;
    end
    Else
      If Abs(y3 - y2) < eps Then
      begin
        m1 := -(x2 - x1) / (y2 - y1);
        mx1 := (x1 + x2) / 2;
        my1 := (y1 + y2) / 2;
        xc := (x3 + x2) / 2;
        yc := m1 * (xc - mx1) + my1;
      end
      Else
      begin
        m1 := -(x2 - x1) / (y2 - y1);
        m2 := -(x3 - x2) / (y3 - y2);
        mx1 := (x1 + x2) / 2;
        mx2 := (x2 + x3) / 2;
        my1 := (y1 + y2) / 2;
        my2 := (y2 + y3) / 2;
        if (m1-m2)<>0 then  //se
        begin
          xc := (m1 * mx1 - m2 * mx2 + my2 - my1) / (m1 - m2);
          yc := m1 * (xc - mx1) + my1;
        end
        else
        begin
          xc:= (x1+x2+x3)/3;
          yc:= (y1+y2+y3)/3;
        end;
      end;

    dx := x2 - xc;
    dy := y2 - yc;
    rsqr := dx * dx + dy * dy;
    r := Sqrt(rsqr);
    dx := xp - xc;
    dy := yp - yc;
    drsqr := dx * dx + dy * dy;

    //store the xc,yc and r for later use
    Triangle^[j].PreCalc:=1;
    Triangle^[j].xc:=xc;
    Triangle^[j].yc:=yc;
    Triangle^[j].r:=r;
  end;

  If drsqr <= rsqr Then InCircle := True;

end;

Function TDelaunay.WhichSide(xp, yp, x1, y1, x2, y2: Double): Integer;
//Determines which side of a line the point (xp,yp) lies.
//The line goes from (x1,y1) to (x2,y2)
//Returns -1 for a point to the left
//         0 for a point on the line
//        +1 for a point to the right
var
 equation: Double;
begin
  equation := ((yp - y1) * (x2 - x1)) - ((y2 - y1) * (xp - x1));
  If equation > 0 Then  WhichSide := -1
  Else
   If equation = 0 Then WhichSide := 0
   Else WhichSide := 1;

End;

Function TDelaunay.Triangulate(nvert: Integer): Integer;
//Takes as input NVERT vertices in arrays Vertex()
//Returned is a list of NTRI triangular faces in the array
//Triangle(). These triangles are arranged in clockwise order.
var

  Complete: PComplete;
  Edges: PEdges;
  Nedge: LongInt;

  //For Super Triangle
  xmin: Double;
  xmax: Double;
  ymin: Double;
  ymax: Double;
  xmid: Double;
  ymid: Double;
  dx: Double;
  dy: Double;
  dz: Double;
  dmax: Double;

  //General Variables
  i : Integer;
  j : Integer;
  k : Integer;
  ntri : Integer;
  xc : Double;
  yc : Double;
  r : Double;
  inc : Boolean;
begin

  //Allocate memory
  GetMem(Complete, sizeof(Complete^));
  GetMem(Edges, sizeof(Edges^));


  //Find the maximum and minimum vertex bounds.
  //This is to allow calculation of the bounding triangle
  xmin := Vertex^[1].x;
  ymin := Vertex^[1].y;
  zmin := Vertex^[1].z;
  xmax := xmin;
  ymax := ymin;
  zmax := zmin;
  For i := 2 To nvert do
  begin
    If Vertex^[i].x < xmin Then xmin := Vertex^[i].x;
    If Vertex^[i].x > xmax Then xmax := Vertex^[i].x;
    If Vertex^[i].y < ymin Then ymin := Vertex^[i].y;
    If Vertex^[i].y > ymax Then ymax := Vertex^[i].y;
    If Vertex^[i].z < zmin Then zmin := Vertex^[i].z;
    If Vertex^[i].z > zmax Then zmax := Vertex^[i].z;
  end;

  dx := xmax - xmin;
  dy := ymax - ymin;
  dz := zmax - zmin;

  If dx > dy Then dmax := dx
  Else dmax := dy;

  xmid := Trunc((xmax + xmin) / 2);
  ymid := Trunc((ymax + ymin) / 2);

  //Set up the supertriangle
  //This is a triangle which encompasses all the sample points.
  //The supertriangle coordinates are added to the end of the
  //vertex list. The supertriangle is the first triangle in
  //the triangle list.

  Vertex^[nvert + 1].x := (xmid - 2 * dmax);
  Vertex^[nvert + 1].y := (ymid - dmax);
  Vertex^[nvert + 2].x := xmid;
  Vertex^[nvert + 2].y := (ymid + 2 * dmax);
  Vertex^[nvert + 3].x := (xmid + 2 * dmax);
  Vertex^[nvert + 3].y := (ymid - dmax);
  Triangle^[1].vv0 := nvert + 1;
  Triangle^[1].vv1 := nvert + 2;
  Triangle^[1].vv2 := nvert + 3;
  Triangle^[1].Precalc := 0;

  Complete[1] := False;
  ntri := 1;

  //Include each point one at a time into the existing mesh
  For i := 1 To nvert do
  begin
    Nedge := 0;
    //Set up the edge buffer.
    //If the point (Vertex(i).x,Vertex(i).y) lies inside the circumcircle then the
    //three edges of that triangle are added to the edge buffer.
    j := 0;
    repeat
      j := j + 1;
      If Complete^[j] <> True Then
      begin
        inc := InCircle(Vertex^[i].x, Vertex^[i].y, Vertex^[Triangle^[j].vv0].x,
                        Vertex^[Triangle^[j].vv0].y, Vertex^[Triangle^[j].vv1].x,
                        Vertex^[Triangle^[j].vv1].y, Vertex^[Triangle^[j].vv2].x,
                        Vertex^[Triangle^[j].vv2].y, xc, yc, r,j);
        //Include this if points are sorted by X
      If (xc + r) < Vertex[i].x Then  //
          complete[j] := True          //
        Else                            //
          If inc Then
          begin
            Edges^[1, Nedge + 1] := Triangle^[j].vv0;
            Edges^[2, Nedge + 1] := Triangle^[j].vv1;
            Edges^[1, Nedge + 2] := Triangle^[j].vv1;
            Edges^[2, Nedge + 2] := Triangle^[j].vv2;
            Edges^[1, Nedge + 3] := Triangle^[j].vv2;
            Edges^[2, Nedge + 3] := Triangle^[j].vv0;
            Nedge := Nedge + 3;
            Triangle^[j].vv0 := Triangle^[ntri].vv0;
            Triangle^[j].vv1 := Triangle^[ntri].vv1;
            Triangle^[j].vv2 := Triangle^[ntri].vv2;
            Triangle^[j].PreCalc:=Triangle^[ntri].PreCalc;
            Triangle^[j].xc:=Triangle^[ntri].xc;
            Triangle^[j].yc:=Triangle^[ntri].yc;
            Triangle^[j].r:=Triangle^[ntri].r;
            Triangle^[ntri].PreCalc:=0;
            Complete^[j] := Complete^[ntri];
            j := j - 1;
            ntri := ntri - 1;
          End;
      End;
    until j>=ntri;

    // Tag multiple edges
    // Note: if all triangles are specified anticlockwise then all
    // interior edges are opposite pointing in direction.
    For j := 1 To Nedge - 1 do
    begin
      If Not (Edges^[1, j] = 0) And Not (Edges^[2, j] = 0) Then
      begin
        For k := j + 1 To Nedge do
        begin
          If Not (Edges^[1, k] = 0) And Not (Edges^[2, k] = 0) Then
          begin
            If Edges^[1, j] = Edges^[2, k] Then
            begin
              If Edges^[2, j] = Edges^[1, k] Then
              begin
                Edges^[1, j] := 0;
                Edges^[2, j] := 0;
                Edges^[1, k] := 0;
                Edges^[2, k] := 0;
              End;
            End;
          End;
        end;
      End;
    end;

    //  Form new triangles for the current point
    //  Skipping over any tagged edges.
    //  All edges are arranged in clockwise order.
    For j := 1 To Nedge do
    begin
      If Not (Edges^[1, j] = 0) And Not (Edges^[2, j] = 0) Then
      begin
        ntri := ntri + 1;
        Triangle^[ntri].vv0 := Edges^[1, j];
        Triangle^[ntri].vv1 := Edges^[2, j];
        Triangle^[ntri].vv2 := i;
        Triangle^[ntri].PreCalc:=0;
        Complete^[ntri] := False;
      End;
    end;
  end;

  //Remove triangles with supertriangle vertices
  //These are triangles which have a vertex number greater than NVERT
  i:= 0;
  repeat
    i := i + 1;
    If (Triangle^[i].vv0 > nvert) Or (Triangle^[i].vv1 > nvert) Or (Triangle^[i].vv2 > nvert) Then
    begin
      Triangle^[i].vv0 := Triangle^[ntri].vv0;
      Triangle^[i].vv1 := Triangle^[ntri].vv1;
      Triangle^[i].vv2 := Triangle^[ntri].vv2;
      i := i - 1;
      ntri := ntri - 1;
    End;
  until i>=ntri;

  Triangulate := ntri;

  //Free memory
  FreeMem(Complete, sizeof(Complete^));
  FreeMem(Edges, sizeof(Edges^));
End;


procedure TDelaunay.Mesh;
begin
  QuickSort(Vertex,1,tPoints-1);
  If tPoints > 3 Then
  HowMany := Triangulate(tPoints-1); //'Returns number of triangles created.
end;


{
procedure TDelaunay.Draw;
var
  //variable to hold how many triangles are created by the triangulate function
  i: Integer;
begin
  // Clear the form canvas
  ClearBackPage;

  TempBuffer.Canvas.Brush.Color := clTeal;
  //Draw the created triangles
  if (HowMany > 0) then
  begin
    For i:= 1 To HowMany do
    begin
    TempBuffer.Canvas.Polygon([Point(Trunc(Vertex^[Triangle^[i].vv0].x), Trunc(Vertex^[Triangle^[i].vv0].y)),
                                  Point(Trunc(Vertex^[Triangle^[i].vv1].x), Trunc(Vertex^[Triangle^[i].vv1].y)),
                                  Point(Trunc(Vertex^[Triangle^[i].vv2].x), Trunc(Vertex^[Triangle^[i].vv2].y))]);
    end;
  end;

  FlipBackPage;
end;
}
procedure TDelaunay.AddPoint(Number:integer;x,y,z:Double);
var
  i: Integer;
  Neu:Boolean;
begin
  //Check for duplicate points
  Neu:=True;
  i:=1;
  while i<tPoints do
  begin
    If (Abs(x-Vertex^[i].x) < ExPtTolerance) and
       (Abs(y-Vertex^[i].y) < ExPtTolerance) Then Neu:=False;
    Inc(i);
  end;

  if Neu then
  begin
//    Vertex^[tPoints].x := Round(x*10)/10;
//    Vertex^[tPoints].y := Round(Y*10)/10;
//    Vertex^[tPoints].z := Round(z*10)/10;
    Vertex^[tPoints].x := Round(x*1E6)/1E6;
    Vertex^[tPoints].y := Round(y*1E6)/1E6;
    Vertex^[tPoints].z := Round(z*1E6)/1E6;
    Vertex^[tpoints].ID:=Number;
    //Increment the total number of points
    inc(tPoints);
  end;
end;
{
procedure TDelaunay.ClearBackPage;
begin
  TempBuffer.Height:=TargetForm.Height;
  TempBuffer.Width:=TargetForm.Width;
  TempBuffer.Canvas.Brush.Color := clSilver;
  TempBuffer.Canvas.FillRect(Rect(0,0,TargetForm.Width,TargetForm.Height));
end;

procedure TDelaunay.FlipBackPage;
var
  ARect : TRect;
begin
  ARect := Rect(0,0,TargetForm.Width,TargetForm.Height);
  TargetForm.Canvas.CopyRect(ARect, TempBuffer.Canvas, ARect);
end;
}

procedure TDelaunay.QuickSort(var A: PVertex; Low,High: Integer);
  //Sort all points by x
  procedure DoQuickSort(var A: PVertex; iLo, iHi: Integer);
  var
    Lo, Hi: Integer;
    Mid: Double;
    T: dVertex;
  begin
    Lo := iLo;
    Hi := iHi;
    Mid := A^[(Lo + Hi) div 2].x;
    repeat
      while A^[Lo].x < Mid do Inc(Lo);
      while A^[Hi].x > Mid do Dec(Hi);
      if Lo <= Hi then
      begin
        T := A^[Lo];
        A^[Lo] := A^[Hi];
        A^[Hi] := T;
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo > Hi;
    if Hi > iLo then DoQuickSort(A, iLo, Hi);
    if Lo < iHi then DoQuickSort(A, Lo, iHi);
  end;
begin
  DoQuickSort(A, Low, High);
end;

function TDelaunay.LineCut(x1,y1,x2,y2,z1,z2,Level: Double): TLineCut;
var
fraction: Double;
begin
 Result.Found:=False;
 if ((z1<=Level) and (z2>Level)) or ((z2<=Level) and (z1>Level)) then
 begin
  fraction:=(z1-Level) / (z1-z2);
  Result.ContourPt.X:=x1 + ((x2-x1)*fraction);
  Result.ContourPt.Y:=y1 + ((y2-y1)*fraction);
  Result.Found:=True;
 end;
end;


{
procedure TDelaunay.DrawContours(canvas:Tcanvas);
var
i,j: Integer;
begin
  Canvas.Pen.Color:=clRed;
  for i:=0 to Contour.NoLevels-1 do
  begin
    for j:=0 to High(Contour.Contours[i]) do
    begin
    Canvas.MoveTo(Round(ConTour.Points[Contour.Contours[i][j].sp].X),Round(ConTour.Points[Contour.Contours[i][j].sp].Y));
    Canvas.LineTo(Round(ConTour.Points[Contour.Contours[i][j].ep].X),Round(ConTour.Points[Contour.Contours[i][j].ep].Y));
    end;
  end;
end;
}

{
procedure TForm1.LoadDataPoints1Click(Sender: TObject);
var
 datapoints: textfile;
 CurrentFile: TFileName;
 numNodes: Integer;
 numTriangles: Integer;
 x,y,z: Double;
 i: Integer;
 no,v1,v2,v3: Integer;
begin

  OpenDialog1.InitialDir:=ExtractFilePath(ParamStr(0));
  if OpenDialog1.Execute then
  begin
    CurrentFile := OpenDialog1.Filename;

  assignfile(datapoints,CurrentFile);
  reset(datapoints);

  readln(datapoints,numNodes);
  SetLength(Nodes,numNodes);

  for i:=0 to numNodes-1 do
  begin
   readln(datapoints,no,x,y,z);
   Nodes[i].X:=Round(x*10000);
   Nodes[i].Y:=Round(y*10000);
   Nodes[i].Z:=Round(z*10000);
  end;

  readln(datapoints,numTriangles);
  SetLength(Triangles,numTriangles);

  for i:=0 to numTriangles-1 do
  begin
   readln(datapoints,no,v1,v2,v3);
   Triangles[i].v1:=v1;
   Triangles[i].v2:=v2;
   Triangles[i].v3:=v3;
  end;

  end;

  Refresh;



end;
}
procedure TDElaunay.InitialiseContourLevels(MinZ,MaxZ,Step:Double);
var
i: Integer;
level: Double;
begin
  if Vertex=NIL then Exit;
  If Step=0 then
  begin
    Step:=abs(MaxZ-MinZ);
    if Step>1000 then Step:=100
    else if Step>500 then Step:=20
    else if Step>100 then Step:=10
    else if Step>50 then Step:=2
    else if Step>10 then Step:=1
    else if Step>1 then Step:=0.1
    else if Step>0.1 then Step:=0.05
    else if Step>0.01 then Step:=0.005
    else if Step>0.001 then Step:=0.0005;
    step:=0.2;
  end;
  Contour.MinValue:=Trunc(MinZ/Step)*Step;
  Contour.MaxValue:=Round(MaxZ/Step)*Step;
  Contour.Range:=Abs(Contour.MaxValue-Contour.MinValue);
  Contour.Step:=Step;
  Contour.NOLevels:=Round(Contour.Range/Contour.Step);
  level:= Contour.MinValue;
  setlength(Contour.levels,Contour.NoLevels);
  SetLength(Contour.Contours,Contour.NoLevels);
  for i:= 0 to Contour.NoLevels-1 do
  begin
    Contour.Levels[i]:=Round(level*1000)/1000;
    level:=level+Contour.Step;
  end;
end;

function TDelaunay.AddContourPoint(P:Tpunkt):Integer;
var i:Integer;
begin
  i:=0;
  while (i<high(Contour.Points)) and
     ((abs(Contour.Points[i].X-P.X)>ExPtTolerance) or (abs(Contour.Points[i].Y-P.Y)>ExPtTolerance)) do inc(i);

  if (length(Contour.Points)=0) or
    ((abs(Contour.Points[i].X-P.X)>ExPtTolerance) or
    (abs(Contour.Points[i].Y-P.Y)>ExPtTolerance)) then
  begin
    setlength(Contour.Points,length(Contour.Points)+1);
    Result:=high(Contour.Points);
    Contour.Points[Result].X:=P.X;
    Contour.Points[Result].Y:=P.Y;
    Contour.Points[Result].ln:=0;
  end
  else Result:=i;
end;

procedure TDelaunay.AddContourLine(Level:Integer;L1,L2:TLineCut);
var i,sp,ep:Integer;
begin
  for i:=0 to high(Contour.Contours[Level]) do
  begin
    If (((abs(L1.ContourPt.x-ConTour.Points[Contour.Contours[Level][i].sp].X)<ExPtTolerance) and
        (abs(L1.ContourPt.y-Contour.Points[Contour.Contours[Level][i].sp].y)<ExPtTolerance))
        or
        ((abs(L1.ContourPt.x-ConTour.Points[Contour.Contours[Level][i].ep].X)<ExPtTolerance) and
         (abs(L1.ContourPt.y-Contour.Points[Contour.Contours[Level][i].ep].y)<ExPtTolerance)))
        and
       (((abs(L2.ContourPt.x-ConTour.Points[Contour.Contours[Level][i].sp].X)<ExPtTolerance) and
        (abs(L2.ContourPt.y-Contour.Points[Contour.Contours[Level][i].sp].y)<ExPtTolerance))
        or
       ((abs(L2.ContourPt.x-ConTour.Points[Contour.Contours[Level][i].ep].X)<ExPtTolerance) and
        (abs(L2.ContourPt.y-Contour.Points[Contour.Contours[Level][i].ep].y)<ExPtTolerance))) then
     EXIT;
  end;

  sp:=AddContourPoint(L1.ContourPt);
  ep:=AddContourPoint(L2.ContourPt);
  if (sp>=0) and (sp<=high(Contour.points)) and
     (ep>=0) and (ep<=high(Contour.points)) then
  begin
    SetLength(Contour.Contours[Level],length(ContOur.Contours[Level])+1);
    i:=high(Contour.Contours[Level]);
    Contour.Contours[Level][i].sp:=sp;
    Contour.Contours[Level][i].ep:=ep;
    Contour.Contours[Level][i].ae:=#0;
  end;
end;

procedure TDelaunay.CalculateContours;
var
line1, line2, line3: TLineCut;
i,j: Integer;
NoLines: Integer;
xa,ya,za,xb,yb,zb,xc,yc,zc: Double;
begin
  for j:=0 to Contour.NoLevels-1 do
  begin
    NoLines:=0;
    for i:=1 to Howmany do
    begin
      xa:=Vertex^[Triangle^[i].vv0].x;
      ya:=Vertex^[Triangle^[i].vv0].y;
      za:=Vertex^[Triangle^[i].vv0].z;
      xb:=Vertex^[Triangle^[i].vv1].x;
      yb:=Vertex^[Triangle^[i].vv1].y;
      zb:=Vertex^[Triangle^[i].vv1].z;
      xc:=Vertex^[Triangle^[i].vv2].x;
      yc:=Vertex^[Triangle^[i].vv2].y;
      zc:=Vertex^[Triangle^[i].vv2].z;

      line1:= LineCut( xa,ya, xb,yb, za,zb, Contour.Levels[j]);
      line2:= LineCut( xb,yb, xc,yc, zb,zc, Contour.Levels[j]);
      line3:= LineCut( xc,yc, xa,ya, zc,za, Contour.Levels[j]);

      if (not line1.Found) then
      begin
        if (line2.Found) and (line3.Found) then
          AddConTourLine(j,Line2,Line3);
      end;

      if (not line2.Found) then
      begin
        if (line1.Found) and (line3.Found) then
          AddConTourLine(j,Line1,Line3);
      end;

      if (not line3.Found) then
      begin
        if (line1.Found) and (line2.Found) then
           AddConTourLine(j,Line1,Line2);
      end;
    end;
  end;
end;

procedure TDelaunay.SortContourLines;
var i,j,la,ia:integer;

procedure SwapAE(var HL:TLine);
var i:integer;
begin
  i:=HL.EP;
  HL.EP:=HL.SP;
  HL.SP:=i;
end;

procedure SwapHLPolY(i,j,k:Integer);
var HL:TLine;
begin
  HL.sp:=Contour.Contours[i][j].sp;
  HL.ep:=Contour.Contours[i][j].ep;
  HL.ae:=Contour.Contours[i][j].ae;
  Contour.Contours[i][j].sp:=Contour.Contours[i][k].sp;
  Contour.Contours[i][j].ep:=Contour.Contours[i][k].ep;
  Contour.Contours[i][j].ae:=Contour.Contours[i][k].ae;
  Contour.Contours[i][k].sp:=HL.sp;
  Contour.Contours[i][k].ep:=HL.ep;
  Contour.Contours[i][k].ae:=HL.ae;
end;

function Endpunkt(k,ep:Integer):Boolean;
var j,n:Integer;
begin
  Result:=False;
  n:=0;
  for j:=Low(Contour.Contours[k]) to high(Contour.Contours[k]) do
    if (Contour.Contours[k][j].sp=ep) or (Contour.Contours[k][j].ep=ep) then
    inc(n);
  Result:=n=1;
end;

function Afangspunkt(k,i:Integer):Boolean;
var j,sp:Integer;
begin
  Result:=False;
  sp:=Contour.Contours[k][i].sp;
  for j:=Low(Contour.Contours[k]) to high(Contour.Contours[k]) do
    if (Contour.Contours[k][j].sp=sp) or (Contour.Contours[k][j].ep=sp) and (j<>i) then
    Result:=True;
end;


begin
 // suche Endpunkte der Hhenline (falls vorhanden)
  for i:=low(Contour.levels) to High(Contour.levels) do
  begin
    for j:=low(Contour.Contours[i]) to high(ConTour.Contours[i]) do
    begin
      If Endpunkt(i,Contour.contours[i][j].sp) then
      begin
        Contour.Contours[i][j].ae:='A';
        Contour.Points[Contour.contours[i][j].sp].ln:=1;
      end;
      If Endpunkt(i,Contour.contours[i][j].ep) then
      begin
        Contour.Contours[i][j].ae:='A';
        Contour.Points[Contour.contours[i][j].ep].ln:=1;
      end;
    end;

    ia:=0;
    for j:=low(Contour.Contours[i]) to high(ConTour.Contours[i]) do
    begin
//      if ((Contour.Points[Contour.Contours[i][j].sp].ln=1) or
//          (Contour.Points[Contour.Contours[i][j].ep].ln=1))  and
      if(Contour.Contours[i][j].ae='A') then
      begin
        SwapHLPOly(i,ia,j);
        inc(ia);
      end;
    end;
    for j:=low(Contour.Contours[i]) to high(ConTour.Contours[i]) do
     If Contour.Contours[i][j].ae<>'A' then Contour.Contours[i][j].ae:=#0;
  end;


  for i:=0 to High(Contour.levels) do
    if length(Contour.Contours[i])>0 then
    begin

      // Offener Linienzug EPA-EPE
      la:=low(contour.Contours[i]);
      repeat
      while (la<high(Contour.Contours[i])) and (Contour.Contours[i][la].ae<>'A') do inc(la);

      If (la<=High(Contour.Contours[i])) and (Contour.Contours[i][la].ae='A') then
      begin
        If Contour.Points[Contour.Contours[i][la].sp].ln<>1 then SwapAE(Contour.Contours[i][la]);
        repeat
          j:=0;
          while (j=la) or  ((j<high(Contour.Contours[i])) and
                (Contour.Contours[i][la].ep<>Contour.Contours[i][j].sp) and
                (Contour.Contours[i][la].ep<>Contour.Contours[i][j].ep)) do inc(j);
          if (j<=high(Contour.Contours[i])) and (j<>la) then
          begin
            if (Contour.Contours[i][la].ep=Contour.Contours[i][j].ep) then SwapAE(Contour.Contours[i][j]);
            inc(la);
            if la<high(Contour.Contours[i]) then SwapHLPOly(i,la,j);
          end;
        until (la>=high(Contour.Contours[i])) or (Contour.Points[Contour.Contours[i][la].ep].ln=1);
        if (la<high(Contour.Contours[i])) and (Contour.Points[Contour.Contours[i][la].ep].ln=1) then
          Contour.Contours[i][la].ae:='E';
      end;

      inc(la);
      until la>high(Contour.Contours[i]);

      la:=High(Contour.Contours[i]);
      while (la>low(Contour.Contours[i])) and (Contour.Contours[i][la].ae=#0) do dec(la);
      inc(la);
      ia:=la;
      if (la>=0) then
      begin
        repeat
          j:=0;
          while (j=la) or ((j<high(Contour.Contours[i])) and
                  (Contour.Contours[i][la].ep<>Contour.Contours[i][j].sp) and
                  (Contour.Contours[i][la].ep<>Contour.Contours[i][j].ep)) do inc(j);
          if (j<=high(Contour.Contours[i])) and (j<>la) and (j<>ia) then
          begin
            if (Contour.Contours[i][la].ep=Contour.Contours[i][j].ep) then SwapAE(Contour.Contours[i][j]);
            if la<high(Contour.Contours[i]) then SwapHLPOly(i,la+1,j);
          end;
          if j=ia then ia:=la+1;

          inc(la);
        until la>high(Contour.Contours[i]);
      end;
    end;
end;

end.
